        /*
 * Copyright (C) 2008 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

 
extern "C" {

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>

}


#include "nativehelper/jni.h"
#include "nativehelper/JNIHelp.h"
#include "images/SkImageDecoder.h"
#include "images/SkImageEncoder.h"
#include "core/SkBitmap.h"
#include "core/SkStream.h"
#include "core/SkTemplates.h"
#include "android_runtime/AndroidRuntime.h"
#include "arcsoft_face_tracking.h"
#include "ammem.h"
#include "merror.h"
#include "arcsoft_face_recognition.h"
#include "arc_redeye_remove.h"
#include "arcsoft_face_beautifier.h"
#include "mdliveeffect.h"
#include "image_codec.h"
#include "amstream.h"

 
#define SAFE_FREE(x)	if (MNull!=(x)) { MMemFree(MNull, (x)); (x) = MNull; }
#define LINE_BYTES16(Width, BitCt) (( ((MLong)(Width) * (BitCt) + 127) >>7)<<4)

static FILE* s_fp = NULL;

void log_to_file(char *fmt,...)
{
    #if 0
	va_list	args;
	char  buf[256];
	if(NULL == s_fp)
		s_fp =fopen("/data/jichen/log_jni.txt", "w");
	if(s_fp != NULL)
	{
		va_start(args, fmt);	
		vsprintf(buf, fmt, args);	
		va_end(args);

		int n = strlen(buf);
		buf[n++] ='\r';
		buf[n++] ='\n';
		fwrite(buf, n, 1, s_fp);
		fflush(s_fp);
	//	fclose(s_fp);
	//	s_fp = NULL;
	}
   #endif
}


static void dump_to_file(const char *fname,

                         uint8_t *buf, uint32_t size)

{

    int nw, cnt = 0;

    uint32_t written = 0;

 

    int fd = open(fname, O_RDWR | O_CREAT);

    if (fd < 0) {

        log_to_file("failed to create file [%s]: %s", fname, strerror(errno));

        return;

    }



    LOGD("writing %d bytes to file [%s]\n", size, fname);

    while (written < size) {

        nw = ::write(fd,

                     buf + written,

                     size - written);

        if (nw < 0) {

            log_to_file("failed to write to file [%s]: %s",

                 fname, strerror(errno));

            break;

        }

        written += nw;

        cnt++;

    }

    

    ::close(fd); 

}

/*
static bool decodeFile(SkBitmap* bitmap, const char srcPath[],SkBitmap::Config config) {
	SkFILEStream stream(srcPath);
	if (!stream.isValid()) {
		SkDebugf("ERROR: bad filename <%s>\n", srcPath);
		return -100;
	}
	SkImageDecoder* codec = SkImageDecoder::Factory(&stream);
	if (NULL == codec) {
		SkDebugf("ERROR: no codec found for <%s>\n", srcPath);
		return -100;
     }
	SkAutoTDelete<SkImageDecoder> ad(codec);
	stream.rewind();
	if (!codec->decode(&stream, bitmap, config, SkImageDecoder::kDecodePixels_Mode)) {
		SkDebugf("ERROR: codec failed for <%s>\n", srcPath);
		return -100;
     }
	return MOK;
}*/
static bool decodeFile(MBITMAP *bitmap, const char srcPath[]) 
{
	MDecode(bitmap,srcPath);
}

/*
static bool decodeFilePartial(SkBitmap* bitmap, const char srcPath[],SkBitmap::Config config,int width,int height) {
	SkBitmap    tmp;
	SkFILEStream stream(srcPath);
	if (!stream.isValid()) {
		SkDebugf("ERROR: bad filename <%s>\n", srcPath);
		return MERR_UNKNOWN;
	}
	SkImageDecoder* codec = SkImageDecoder::Factory(&stream);
	if (NULL == codec) {
		SkDebugf("ERROR: no codec found for <%s>\n", srcPath);
		return MERR_UNKNOWN;
    }
	SkAutoTDelete<SkImageDecoder> ad(codec);
	stream.rewind();
	tmp.setConfig(config, width, height);
	if (!codec->decode(&stream, &tmp, config, SkImageDecoder::kDecodePixels_Mode)) {
		SkDebugf("ERROR: codec failed for <%s>\n", srcPath);
		return MERR_UNKNOWN;
    }
	bitmap->swap(tmp);
	return MOK;
}
 
 
static bool encodeFile(const char file[], const SkBitmap& bm, SkImageEncoder::Type t,
                                int quality)
{
    SkAutoTDelete<SkImageEncoder> enc(SkImageEncoder::Create(t));
    if ((enc.get() && enc.get()->encodeFile(file, bm, quality) ) != true)
		return MERR_UNKNOWN;

	return MOK;
}*/

static bool encodeFile(const char file[], MBITMAP *bm,const char thumbnaifile[],const MBITMAP *thumbnaibitmap,MLong quality)
{
	MEncode(file,bm,MNull,MNull,quality);
}
 
static MRESULT multi_area(MBITMAP *Dst,MBITMAP *Src,MRECT *rect,MLong num)
{
	MByte *pdst = MNull,*psrc = MNull;
	MLong i = 0,j = 0,k = 0;
	MLong lalphatemp = (MLong)(0.4*256);
	MLong lwidth = 0,lheight = 0;
	MByte RR,GG,BB;

	lwidth  = Dst->lWidth;
	lheight = Dst->lHeight;

	for (i = 0;i < lheight;i++)
	{
		pdst = Dst->pPlane[0] + i * Dst->lPitch[0];
		psrc = Src->pPlane[0] + i * Src->lPitch[0];
		for (j = 0;j < lwidth;j++)
		{
			MWord pSrctemp;
			pSrctemp = *((MWord*)psrc);
			RR = (pSrctemp >> 8) & 0xF8;
			GG = (pSrctemp >> 3) & 0xFC;
			BB = (pSrctemp << 3) & 0xF8;

			RR = (MByte)((RR * lalphatemp)>>8);
			GG = (MByte)((GG * lalphatemp)>>8);
			BB = (MByte)((BB * lalphatemp)>>8);

			for (k = 0;k < num;k++)
			{
				if (i >= rect[k].top && i < rect[k].bottom && j >= rect[k].left && j < rect[k].right)
				{
					RR = (pSrctemp >> 8) & 0xF8;
					GG = (pSrctemp >> 3) & 0xFC;
					BB = (pSrctemp << 3) & 0xF8;
				}
			}

			*((MWord *)pdst) = (RR >> 3 << 11) | (GG >> 2 << 5) | (BB >> 3);

			psrc += 2;
			pdst += 2;
		}
	}
	return MOK;
}

static MVoid GetAlpha(MByte *pDst8,MByte *pSrc32,MLong lWidth,MLong lHeight,MLong lDstpitch,MLong lSrcpitch)
{
	MLong width = 0,height = 0,i = 0,j = 0,Srcpitch = lSrcpitch,Dstpitch = lDstpitch;
	MByte *pDst,*pSrc;
	
	width  = lWidth;
	height = lHeight;
	
	for(i = 0;i < height;i++)
	{
		pSrc = pSrc32 + i * Srcpitch;
		pDst = pDst8  + i * lDstpitch;
		for(j = 0;j < width;j++)
		{
			*pDst++ = pSrc[3];
			pSrc += 4;
		}
	}
}


static jint Java_com_arcsoft_arcsoftservice_ArcSoftService_FaceDetect(JNIEnv *env, jobject thiz, 
									    jbyteArray imgDatra, jint width,jint linebyte, jint height, jint maxFaceNum, jint Priority, 
										jintArray FaceRes, jintArray FaceNum) {
	int result = 1000;

     int *pFaceNum = (int *)env->GetPrimitiveArrayCritical(FaceNum,0);
    int *pFaceRes = (int *)env->GetPrimitiveArrayCritical(FaceRes,0);
    unsigned char *pData = (unsigned char *)env->GetPrimitiveArrayCritical(imgDatra,0);


 //	*pFaceNum = 6666;
#define FT_MEM	(1.5*1024*1024)
	//do face detection
    
   	AMFPAF_IMAGEINFO	imageInfo={0};
	imageInfo.lWidth  = width;
	imageInfo.lHeight = height;

	imageInfo.lPixelArrayFormat = AMFPAF_PAF_R5G6B5;
	imageInfo.lLineBytes = linebyte; //width*2;

	MVoid*	pMemBuff = MMemAlloc(MNull, FT_MEM);
	if(pMemBuff == 0)
       return -1;

	MHandle			hMemMgr = MMemMgrCreate(pMemBuff, FT_MEM);  
	AMFPAF_ENGINE	faceEngine;
	AMFPAF_FACERES	faceRes;

	result = AMFPAF_InitialFaceEngine(hMemMgr, &imageInfo, &faceRes, &faceEngine, \
			Priority, 16, maxFaceNum);

	if(result != MERR_NONE){
		SAFE_FREE(pMemBuff);
		env->ReleasePrimitiveArrayCritical(FaceNum,pFaceNum,0);
		env->ReleasePrimitiveArrayCritical(FaceRes,pFaceRes,0);
		env->ReleasePrimitiveArrayCritical(imgDatra,pData,0);
		return -1;
	}

	result = AMFPAF_StillImageFaceDetection(			        // return MOK if success, otherwise fail
	  			hMemMgr,		                // [in]  User defined memory manager
	 		    faceEngine,		                // [in]	 Detect engine/
	            (AMFPAF_PIXELARRAY)pData,		                // [in]  The original image data
	            &faceRes,		                // [out] The detection result
	            &imageInfo,						// [in]  The information of input image
	            MNull,	        // [in]  The callback function
	            MNull,		// [in]  Caller-specific data that will be passed into the callback function
	            MNull,	                    // [in]  The callback function
	            MNull		            // [in]  Caller-specific data that will be passed into the callback function
);

//	dump_to_file("/data/jichen/face.dat", (uint8_t *)pData, height*width*2);

	if(result == 0)  
	{


        log_to_file("Detect %d face!!!!!!!!",faceRes.nFace );
	 	*pFaceNum = faceRes.nFace;
	//	env->SetIntArrayRegion(FaceNum, 0, 1, (const int *)&faceRes.nFace);
		
		for(int i = 0; i<faceRes.nFace; i++)
		{
			pFaceRes[5*i] = faceRes.rcFace[i].top;
			pFaceRes[5*i+1] = faceRes.rcFace[i].bottom;
			pFaceRes[5*i+2] = faceRes.rcFace[i].left;
			pFaceRes[5*i+3] = faceRes.rcFace[i].right;
			pFaceRes[5*i+4] = faceRes.lfaceOrient;
		}
	}

       
	
    	SAFE_FREE(pMemBuff);
 
   log_to_file("pFaceNum[0] %d!!!!!!!!",faceRes.nFace );
 //end
 	env->ReleasePrimitiveArrayCritical(FaceNum,pFaceNum,0);
	env->ReleasePrimitiveArrayCritical(FaceRes,pFaceRes,0);
	env->ReleasePrimitiveArrayCritical(imgDatra,pData,0);
 	return result;

 
}

 

static jint Java_com_arcsoft_arcsoftservice_ArcSoftService_FaceRegister(JNIEnv *env, jobject thiz, 
									    jbyteArray imgDatra, jint register , jint width , jint linebyte, jint height, 
										jintArray FaceRes, jbyteArray FRFeature, jintArray Featuresize) {
	int result = 1000;

    unsigned char * pFRFeature = (unsigned char *)env->GetPrimitiveArrayCritical(FRFeature,0);
    int *pFaceRes = (int *)env->GetPrimitiveArrayCritical(FaceRes,0);
    unsigned char *pData = (unsigned char *)env->GetPrimitiveArrayCritical(imgDatra,0);
 	int * pFeaturesize = (int *)env->GetPrimitiveArrayCritical(Featuresize,0);
#define FR_MEM	(1887436)//(1.8*1024*1024)

    MHandle		FR_Engine;
	MVoid* pMemBuff = MMemAlloc(MNull, FR_MEM);
	if (MNull==pMemBuff)
	{
		env->ReleasePrimitiveArrayCritical(FRFeature,pFRFeature,0);
		env->ReleasePrimitiveArrayCritical(FaceRes,pFaceRes,0);
		env->ReleasePrimitiveArrayCritical(imgDatra,pData,0);
		env->ReleasePrimitiveArrayCritical(Featuresize, pFeaturesize, 0);
		return -1;
	}
 
	MHandle	FR_hMemMgr= MMemMgrCreate(pMemBuff, FR_MEM); 
		 	// initialize the FR engine 		
		result = AFFR_InitialEngine(FR_hMemMgr, &FR_Engine);
		if (MERR_NONE != result)
		{
			log_to_file("FR AFFR_InitialEngine error\n");
			SAFE_FREE(pMemBuff);
			env->ReleasePrimitiveArrayCritical(FRFeature,pFRFeature,0);
			env->ReleasePrimitiveArrayCritical(FaceRes,pFaceRes,0);
			env->ReleasePrimitiveArrayCritical(imgDatra,pData,0);
			env->ReleasePrimitiveArrayCritical(Featuresize, pFeaturesize, 0);
			return -1;
		}
	 
	 AFFR_OFFSCREEN inputImage;
	 AFFR_FACIALPOINTS facialPoints;
     AFFR_FeatureData featureData;
 	 AFFR_FRFeature reffeature;
	 AFFR_FACEMODELS facemodels;
	 	  
     MRECT face;
	 MLong  FaceOrient;
	 face.top = pFaceRes[0];
	 face.bottom = pFaceRes[1];
	 face.left = pFaceRes[2];
	 face.right = pFaceRes[3];
	 FaceOrient = pFaceRes[4];

	 AFFR_FACEINPUT faceRes;

	 faceRes.pFaceRectArray = &face;
	 faceRes.pFaceOrientArray =  &FaceOrient;
	 faceRes.lFaceNumber = 1;

	 inputImage.lWidth = width;
	 inputImage.lHeight = height;
	 inputImage.lPixelArrayFormat = AFFR_PAF_R5G6B5;
	 inputImage.pixelArray.chunky.lLineBytes = linebyte; //width*2;
	 inputImage.pixelArray.chunky.pPixel = (MByte*)pData;



 	  
	 result = AFFR_FacialFeatureLocation(FR_hMemMgr, FR_Engine, &inputImage, &faceRes, MTrue, &facialPoints);
	 if (result != MOK)
	 {
		 SAFE_FREE(pMemBuff);
		 log_to_file("AFFR_FacialFeatureLocation error\n");
		 env->ReleasePrimitiveArrayCritical(FRFeature,pFRFeature,0);
		 env->ReleasePrimitiveArrayCritical(FaceRes,pFaceRes,0);
		 env->ReleasePrimitiveArrayCritical(imgDatra,pData,0);
		 env->ReleasePrimitiveArrayCritical(Featuresize, pFeaturesize, 0);
		 return result;
	 }

	 if (!(facialPoints.pFlagArray[0]))
		{
		    	SAFE_FREE(pMemBuff);
				log_to_file("facialPoints.pFlagArray[0] error\n");
				env->ReleasePrimitiveArrayCritical(FRFeature,pFRFeature,0);
				env->ReleasePrimitiveArrayCritical(FaceRes,pFaceRes,0);
				env->ReleasePrimitiveArrayCritical(imgDatra,pData,0);
				env->ReleasePrimitiveArrayCritical(Featuresize, pFeaturesize, 0);
				return -1;
		}

	// featureData = facialPoints.pFeatureDataArray[0];

	 result = AFFR_ExtractFRFeature(FR_hMemMgr, FR_Engine, &inputImage, &facialPoints, MTrue, &facemodels);
		if (result != MOK)
		{
			SAFE_FREE(pMemBuff);
			log_to_file("AFFR_ExtractFRFeature error\n");
			env->ReleasePrimitiveArrayCritical(FRFeature,pFRFeature,0);
			env->ReleasePrimitiveArrayCritical(FaceRes,pFaceRes,0);
			env->ReleasePrimitiveArrayCritical(imgDatra,pData,0);
			env->ReleasePrimitiveArrayCritical(Featuresize, pFeaturesize, 0);
			return result;
		 
		}
 
		
		if(facemodels.lModelNumber != 0){

			//ֻ֧һԿôд
			memcpy(pFRFeature,facemodels.pFaceModelArray[0].pfeature, facemodels.pFaceModelArray[0].buffersize);
		 	*pFeaturesize = facemodels.pFaceModelArray[0].buffersize;
         //   env->SetIntArrayRegion(Featuresize, 0, 1, (const int *)&(facemodels.pFaceModelArray[0].buffersize));
		//	log_to_file("*pFeaturesize = %d\n, ", *pFeaturesize);
			
		}else
		{
			// no feature
			log_to_file("AFFR_ExtractFRFeature get NO feature\n");
			env->ReleasePrimitiveArrayCritical(FRFeature,pFRFeature,0);
			env->ReleasePrimitiveArrayCritical(FaceRes,pFaceRes,0);
			env->ReleasePrimitiveArrayCritical(imgDatra,pData,0);
			env->ReleasePrimitiveArrayCritical(Featuresize, pFeaturesize, 0);
			result = -1;
		}

	SAFE_FREE(pMemBuff);
	env->ReleasePrimitiveArrayCritical(FRFeature,pFRFeature,0);
	env->ReleasePrimitiveArrayCritical(FaceRes,pFaceRes,0);
	env->ReleasePrimitiveArrayCritical(imgDatra,pData,0);
 	env->ReleasePrimitiveArrayCritical(Featuresize, pFeaturesize, 0);
 	return result;

}



static jint Java_com_arcsoft_arcsoftservice_ArcSoftService_FaceRegonition(JNIEnv *env, jobject thiz, 
									    jbyteArray reffeature, jint refNum, jbyteArray probfeature, 
										jintArray frResult_flag, jintArray frResult_score, jintArray pIID) {
	int result = 1000;

    unsigned char * preffeature = (unsigned char *)env->GetPrimitiveArrayCritical(reffeature,0);
     int *pfrResult_flag = (int *)env->GetPrimitiveArrayCritical(frResult_flag,0);
    unsigned char *pprobfeature = (unsigned char *)env->GetPrimitiveArrayCritical(probfeature,0);
 	int * pfrResult_score = (int *)env->GetPrimitiveArrayCritical(frResult_score,0);
 	int * ppIID = (int *)env->GetPrimitiveArrayCritical(pIID,0);


	MHandle		FR_Engine;
	MVoid* pMemBuff = MMemAlloc(MNull, FR_MEM);
	if (MNull==pMemBuff)
	{
		env->ReleasePrimitiveArrayCritical(reffeature,preffeature,0);
		env->ReleasePrimitiveArrayCritical(frResult_flag,pfrResult_flag,0);
		env->ReleasePrimitiveArrayCritical(probfeature,pprobfeature,0);
		env->ReleasePrimitiveArrayCritical(frResult_score, pfrResult_score, 0);
		env->ReleasePrimitiveArrayCritical(pIID, ppIID, 0);
		return -1;
	}
	
	MHandle	FR_hMemMgr= MMemMgrCreate(pMemBuff, FR_MEM); 
		 	// initialize the FR engine 		
	result = AFFR_InitialEngine(FR_hMemMgr, &FR_Engine);
	if (MERR_NONE != result)
	{
		log_to_file("FR AFFR_InitialEngine error\n");
		SAFE_FREE(pMemBuff);
		env->ReleasePrimitiveArrayCritical(reffeature,preffeature,0);
		env->ReleasePrimitiveArrayCritical(frResult_flag,pfrResult_flag,0);
		env->ReleasePrimitiveArrayCritical(probfeature,pprobfeature,0);
		env->ReleasePrimitiveArrayCritical(frResult_score, pfrResult_score, 0);
		env->ReleasePrimitiveArrayCritical(pIID, ppIID, 0);
		return -1;
	}

	AFFR_FRFeature *RefFeatures = (AFFR_FRFeature *)MMemAlloc(MNull, sizeof(AFFR_FRFeature)*refNum);

 
		 for(int i=0; i<refNum; i++){
			 RefFeatures[i].buffersize = 1200;
			 RefFeatures[i].pfeature = preffeature + i*RefFeatures[i].buffersize;
		 }
	 

    AFFR_FRFeature probefeature;
    probefeature.buffersize = 1200;
    probefeature.pfeature = pprobfeature;
    

	AFFR_Result frResult;
	MLong max_idx;


	result = AFFR_AuthenticationEx(FR_hMemMgr, FR_Engine, 
			RefFeatures, refNum, &probefeature, &frResult, &max_idx);
	if (result != MOK){
		log_to_file("FR AFFR_AuthenticationEx error, result =%d\n", result);
		SAFE_FREE(pMemBuff);
		SAFE_FREE(RefFeatures);
		env->ReleasePrimitiveArrayCritical(reffeature,preffeature,0);
		env->ReleasePrimitiveArrayCritical(frResult_flag,pfrResult_flag,0);
		env->ReleasePrimitiveArrayCritical(probfeature,pprobfeature,0);
		env->ReleasePrimitiveArrayCritical(frResult_score, pfrResult_score, 0);
		env->ReleasePrimitiveArrayCritical(pIID, ppIID, 0);
		return result;
	}

 	*pfrResult_flag = frResult.flag;
     *pfrResult_score = (int)(frResult.score*100);
 	*ppIID = max_idx;  
//    jint flag = frResult.flag;
//	env->SetIntArrayRegion(frResult_flag, 0, 1, &(flag));
//	jint score = (int)(frResult.score*100);
//	env->SetIntArrayRegion(frResult_score, 0, 1, (const int *)&(score));
//	env->SetIntArrayRegion(pIID, 0, 1, (const int *)&(max_idx));

	if(frResult.flag == 1)
       	log_to_file("FR find people\n");
	else{
       	log_to_file("No people, refNum = %d\n", refNum);

	    log_to_file("preffeature[0]=%d, preffeature[1]=%d, preffeature[2]=%d, preffeature[3]=%d\n",
			preffeature[0], preffeature[1], preffeature[2], preffeature[3] );

        log_to_file("pprobfeature[0]=%d, pprobfeature[1]=%d, pprobfeature2]=%d, pprobfeature[3]=%d\n",
			pprobfeature[0], pprobfeature[1], pprobfeature[2], pprobfeature[3] );

//	    dump_to_file("/data/jichen/feature_old.dat", (uint8_t *)preffeature, 1200);
// 		 dump_to_file("/data/jichen/feature.dat", (uint8_t *)pprobfeature, 1200);
	
	}

//	  	log_to_file("frResult.score = %f, *pfrResult_score =%d\n", frResult.score, *pfrResult_score);


	SAFE_FREE(pMemBuff);
	SAFE_FREE(RefFeatures);
	env->ReleasePrimitiveArrayCritical(reffeature,preffeature,0);
 	env->ReleasePrimitiveArrayCritical(frResult_flag,pfrResult_flag,0);
	env->ReleasePrimitiveArrayCritical(probfeature,pprobfeature,0);
 	env->ReleasePrimitiveArrayCritical(frResult_score, pfrResult_score, 0);
 	env->ReleasePrimitiveArrayCritical(pIID, ppIID, 0);
 	return result;

}


static int Do_RER( unsigned char * pdata , int width , int height, int linebyte)
{
 
	int result = 1000;

	ARER_OFFSCREEN RERImage;
	unsigned long nRERWorkMemSize = 300*1024;
	MByte *pRERWorkMem;
 

#define REDEYE_MAX 200

	pRERWorkMem = (MByte *)MMemAlloc (MNull, nRERWorkMemSize);
	if (MNull==pRERWorkMem)
	{
		return -1;
	}
	
 

   RERImage.lPixelArrayFormat = ARER_PAF_RGB565;
   RERImage.lWidth = width;
   RERImage.lHeight = height;
   RERImage.pixelArray.chunky.lLineBytes = linebyte;
   RERImage.pixelArray.chunky.pPixel = pdata;

   int redeye_num;

   redeye_num = ARER_AutoRemove (
	&RERImage,
	pRERWorkMem,
	nRERWorkMemSize,
	MNull,
	MNull);
 
     log_to_file(" find %d eys\n",redeye_num );
	
   
    	SAFE_FREE(pRERWorkMem);
 
	  	return redeye_num;

}

static jint Java_com_arcsoft_arcsoftservice_ArcSoftService_RER(JNIEnv *env, jobject thiz, 
									    jbyteArray imgDatra , jint width, jint linebyte, jint height)
{
	unsigned char * pdata = (unsigned char *)env->GetPrimitiveArrayCritical(imgDatra,0);
	int redeye_num;

	redeye_num = Do_RER(pdata, width, height,linebyte);

	 
	env->ReleasePrimitiveArrayCritical(imgDatra,pdata,0);

	  	return redeye_num;

}


static jint do_FB( unsigned char * pdata , int width , int height, int linebyte, 
				  int option, int cleanLevel, int *pFaceRes, int FaceNum)
{
 
    int result = 1000;
	
#define FB_BUFFER				1024*1024*3/2	


    MVoid *pWorkBuffer = MMemAlloc (MNull,FB_BUFFER);
	if (MNull==pWorkBuffer)
	{
		return -1;
	}
	
	MHandle hMemMgr = MMemMgrCreate(pWorkBuffer, FB_BUFFER);
	MHandle hBeautifier = MNull;
	if(0 !=(result = AFB_Init(hMemMgr, &hBeautifier)))
         log_to_file(" AFB_Beautify error,return %d \n",result );
	if(0 !=(result =AFB_SetCleanModel(hBeautifier, option)))
         log_to_file(" AFB_SetCleanModel error,return %d \n",result );
	AFB_SetCleanLevel(hBeautifier, cleanLevel);

	AFB_OFFSCREEN img = {0};  
	img.lWidth = width;
	img.lHeight = height;
	img.lPixelArrayFormat = AFB_PAF_RGB565;
	img.pixelArray.chunky.lLineBytes = linebyte; 
	img.pixelArray.chunky.pPixel = pdata;


	MRECT *prtFaces = (MRECT *)MMemAlloc (MNull,FaceNum*sizeof(MRECT));
	if (MNull==prtFaces)
	{
		SAFE_FREE(pWorkBuffer);
		return -1;
	}
	for(int j=0; j<FaceNum; j++)
	{
		prtFaces[j].top	= pFaceRes[j*5+0];
		prtFaces[j].bottom	= pFaceRes[j*5+1];
		prtFaces[j].left	= pFaceRes[j*5+2];
		prtFaces[j].right	=  pFaceRes[j*5+3];

	}

	AFB_FACES arrFaces = {0}; 
	arrFaces.prtFaces = prtFaces;
	arrFaces.lFacesNum = FaceNum;


//	log_to_file(" arrFaces.prtFaces[0].top = %d \n",arrFaces.prtFaces[0].top );
//	log_to_file(" arrFaces.prtFaces[0].bottom = %d \n",arrFaces.prtFaces[0].bottom );
//	log_to_file(" arrFaces.prtFaces[0].left = %d \n",arrFaces.prtFaces[0].left );
//	log_to_file(" arrFaces.prtFaces[0].right = %d \n",arrFaces.prtFaces[0].right );

	if(0 != (result = AFB_Beautify(hBeautifier, &img, &arrFaces, &img, MNull, MNull)))
         log_to_file(" AFB_Beautify error,return %d \n",result );
	
//	dump_to_file("/data/jichen/face.dat", (uint8_t *)pdata, height*width*2);
	
	SAFE_FREE(pWorkBuffer);
	SAFE_FREE(prtFaces);
 
	return result;
}



static jint Java_com_arcsoft_arcsoftservice_ArcSoftService_FB(JNIEnv *env, jobject thiz, 
									    jbyteArray imgDatra , jint width , jint linebyte, jint height,
										jint option, jint cleanLevel, jintArray FaceRes, jint FaceNum)
{
	unsigned char * pdata = (unsigned char *)env->GetPrimitiveArrayCritical(imgDatra,0);
	int *pFaceRes = (int *)env->GetPrimitiveArrayCritical(FaceRes,0);
    int result = do_FB(pdata ,  width , height, linebyte,   
				   option,  cleanLevel, pFaceRes, FaceNum);
	env->ReleasePrimitiveArrayCritical(imgDatra,pdata,0);
	env->ReleasePrimitiveArrayCritical(FaceRes,pFaceRes,0);
	
	return result;
}


static jint Java_com_arcsoft_arcsoftservice_ArcSoftService_LiveEffect(JNIEnv *env, jobject c,jbyteArray imgData,jint width, jint height,jint linebyte,
								jint effectID, jintArray jpara,jint dwparamnum)
{
	jobject 	obj,obj1;
	jfieldID	fieldId;
	jclass      clazz;
    int res = MOK;
	MHandle phLiveEffect;
	SPOTLIGHTPARAM   spot_para;
	MOSAICPARAM      mosaic_para;
	WarpPinchPARAM   warp_para;
	BigEyePARAM      bigeye_para;
	BLURPARAM        blur_para;
	int lSize = 0;
	MRECT DstRect,tempRect;
	MRECT *ProcessRect;
	MLong i = 0;
	MBITMAP *bitmap;
	int *para = (int *)env->GetPrimitiveArrayCritical(jpara,0);
	int *pData = (int *)env->GetPrimitiveArrayCritical(imgData,0);

	bitmap = (MBITMAP *)MMemAlloc(MNull,sizeof(MBITMAP));//MERR_NO_MEMORY
	if(!bitmap)
	{
		res = MERR_NO_MEMORY;
		goto EXIT;
	}
	bitmap->dwPixelArrayFormat = MPAF_RGB16_R5G6B5;
	bitmap->lWidth          = width;
	bitmap->lHeight         = height;
	bitmap->lPitch[0]       = linebyte;
	bitmap->lPitch[1]       = 0;
	bitmap->lPitch[2]       = 0;
	bitmap->pPlane[0]       = (MByte *)pData;
	bitmap->pPlane[1]       = MNull;
	bitmap->pPlane[2]       = MNull;

	DstRect.left = 0;
	DstRect.right = width;
	DstRect.top = 0;
	DstRect.bottom = height;

	if (effectID == MLIVEEFFECT_ID_SPOTLIGHT)//spotlight
	{
		lSize = sizeof(SPOTLIGHTPARAM);
		ProcessRect = (MRECT *)MMemAlloc(MNull,sizeof(MRECT) * dwparamnum);
		for (i = 0; i < dwparamnum;i++)
		{
			ProcessRect[i].left = para[(i << 2)] - para[(i << 2) + 2];
			ProcessRect[i].right = para[(i << 2)] + para[(i << 2) + 2];
			ProcessRect[i].top  = para[(i << 2) + 1] - para[(i << 2) + 3];
			ProcessRect[i].bottom = para[(i << 2) + 1] + para[(i << 2) + 3];
		}
		for (i = 0;i < dwparamnum;i++)
		{
			spot_para.lcircleCentX = para[(i << 2)];
			spot_para.lcircleCentY = para[(i << 2) + 1];
			spot_para.lcircleRadiusx = para[(i << 2) + 2];
			spot_para.lcircleRadiusy = para[(i << 2) + 3];

			tempRect.left = ProcessRect[i].left;
			tempRect.right = ProcessRect[i].right;
			tempRect.top   = ProcessRect[i].top;
			tempRect.bottom = ProcessRect[i].bottom;

			res = MLiveEffect_Create(width,height,effectID,&spot_para,lSize,&phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Create");
				goto EXIT;
			}

			res = MLiveEffect_Begin(phLiveEffect, width, height, &tempRect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Begin");
				goto EXIT;
			}

			res = MLiveEffect_Process(phLiveEffect,&DstRect,bitmap,&DstRect,bitmap);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Process");
				goto EXIT;
			}

			res = MLiveEffect_End(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_End");
				goto EXIT;
			}

			res = MLiveEffect_Destroy(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Destroy");
				goto EXIT;
			}
		}

		multi_area(bitmap,bitmap,ProcessRect,dwparamnum);
		
		if (ProcessRect)
		{
			MMemFree(MNull,ProcessRect);
		}
	}
	else if (effectID == MLIVEEFFECT_ID_MOSAIC)//mosaic
	{
		lSize  = sizeof(MOSAICPARAM);
		for (i = 0;i < dwparamnum;i++)
		{
			tempRect.left  = para[i * 5];
			tempRect.right = para[i * 5 + 1];
			tempRect.top   = para[i * 5 + 2];
			tempRect.bottom= para[i * 5 + 3];
			mosaic_para.lMosaicSize = para[i * 5 + 4];
			
			res = MLiveEffect_Create(width,height,effectID,&mosaic_para,lSize,&phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Create");
				goto EXIT;
			}
				
			res = MLiveEffect_Begin(phLiveEffect, width,height, &tempRect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Begin");
				goto EXIT;
			}
				
			res = MLiveEffect_Process(phLiveEffect,&DstRect,bitmap,&DstRect,bitmap);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Process");
				goto EXIT;
			}
			res = MLiveEffect_End(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_End");
				goto EXIT;
			}
			res = MLiveEffect_Destroy(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Destroy");
				goto EXIT;
			}
		}
	}
	else if (effectID == MLIVEEFFECT_ID_WARP || effectID == MLIVEEFFECT_ID_PINCH)//warp or pinch
	{
		lSize = sizeof(WarpPinchPARAM);
		ProcessRect = (MRECT *)MMemAlloc(MNull,sizeof(MRECT) * dwparamnum);
		for (i = 0; i < dwparamnum;i++)
		{
			ProcessRect[i].left = para[(i * 5)] - para[(i * 5) + 2];
			ProcessRect[i].right = para[(i * 5)] + para[(i * 5) + 2];
			ProcessRect[i].top  = para[(i * 5) + 1] - para[(i * 5) + 3];
			ProcessRect[i].bottom = para[(i * 5) + 1] + para[(i * 5) + 3];
		}
		for (i = 0;i < dwparamnum;i++)
		{
			warp_para.lcircleCentX = para[i * 5 ];
			warp_para.lcircleCentY = para[i * 5 + 1];
			warp_para.lcircleRadiusx = para[i * 5 + 2];
			warp_para.lcircleRadiusy = para[i * 5 + 3];
			warp_para.lWarpPinchSize = para[i * 5 + 4];

			tempRect.left = ProcessRect[i].left;
			tempRect.right = ProcessRect[i].right;
			tempRect.top   = ProcessRect[i].top;
			tempRect.bottom = ProcessRect[i].bottom;

			res = MLiveEffect_Create(width,height,effectID,&warp_para,lSize,&phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Create");
				goto EXIT;
			}

			res = MLiveEffect_Begin(phLiveEffect, width, height, &tempRect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Begin");
				goto EXIT;
			}

			res = MLiveEffect_Process(phLiveEffect,&DstRect,bitmap,&DstRect,bitmap);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Process");
				goto EXIT;
			}

			res = MLiveEffect_End(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_End");
				goto EXIT;
			}

			res = MLiveEffect_Destroy(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Destroy");
				goto EXIT;
			}
		}
		if (ProcessRect)
		{
			MMemFree(MNull,ProcessRect);
		}
	}
	else if (effectID == MLIVEEFFECT_ID_BIGEYE)
	{
		lSize = sizeof(BigEyePARAM);
		ProcessRect = (MRECT *)MMemAlloc(MNull,sizeof(MRECT) * dwparamnum);
		for (i = 0; i < dwparamnum;i++)
		{
			ProcessRect[i].left = para[(i * 5)] - para[(i * 5) + 2];
			ProcessRect[i].right = para[(i * 5)] + para[(i * 5) + 2];
			ProcessRect[i].top  = para[(i * 5) + 1] - para[(i * 5) + 3];
			ProcessRect[i].bottom = para[(i * 5) + 1] + para[(i * 5) + 3];
		}
		for (i = 0;i < dwparamnum;i++)
		{
			bigeye_para.lcircleCentX = para[i * 5 ];
			bigeye_para.lcircleCentY = para[i * 5 + 1];
			bigeye_para.lcircleRadiusx = para[i * 5 + 2];
			bigeye_para.lcircleRadiusy = para[i * 5 + 3];
			bigeye_para.lBigEyeSize = para[i * 5 + 4];

			tempRect.left = ProcessRect[i].left;
			tempRect.right = ProcessRect[i].right;
			tempRect.top   = ProcessRect[i].top;
			tempRect.bottom = ProcessRect[i].bottom;

			res = MLiveEffect_Create(width,height,effectID,&bigeye_para,lSize,&phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Create");
				goto EXIT;
			}

			res = MLiveEffect_Begin(phLiveEffect, width, height, &tempRect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Begin");
				goto EXIT;
			}

			res = MLiveEffect_Process(phLiveEffect,&DstRect,bitmap,&DstRect,bitmap);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Process");
				goto EXIT;
			}

			res = MLiveEffect_End(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_End");
				goto EXIT;
			}

			res = MLiveEffect_Destroy(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Destroy");
				goto EXIT;
			}
		}
		if (ProcessRect)
		{
			MMemFree(MNull,ProcessRect);
		}
	}
	else if (effectID == MLIVEEFFECT_ID_BLUR)//only support one rect
	{
		lSize = sizeof(BLURPARAM);
		dwparamnum = 1;
		ProcessRect = (MRECT *)MMemAlloc(MNull,sizeof(MRECT) * dwparamnum);
		for (i = 0; i < dwparamnum;i++)
		{
			ProcessRect[i].left = para[(i * 5)] - para[(i * 5) + 2];
			ProcessRect[i].right = para[(i * 5)] + para[(i * 5) + 2];
			ProcessRect[i].top  = para[(i * 5) + 1] - para[(i * 5) + 3];
			ProcessRect[i].bottom = para[(i * 5) + 1] + para[(i * 5) + 3];
		}
		for (i = 0;i < dwparamnum;i++)
		{
			blur_para.lcircleCentX = para[i * 5 ];
			blur_para.lcircleCentY = para[i * 5 + 1];
			blur_para.lcircleRadiusx = para[i * 5 + 2];
			blur_para.lcircleRadiusy = para[i * 5 + 3];
			blur_para.lBlurSize = para[i * 5 + 4];

			tempRect.left = ProcessRect[i].left;
			tempRect.right = ProcessRect[i].right;
			tempRect.top   = ProcessRect[i].top;
			tempRect.bottom = ProcessRect[i].bottom;

			res = MLiveEffect_Create(width,height,effectID,&blur_para,lSize,&phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Create");
				goto EXIT;
			}

			res = MLiveEffect_Begin(phLiveEffect, width, height, &DstRect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Begin");
				goto EXIT;
			}

			res = MLiveEffect_Process(phLiveEffect,&DstRect,bitmap,&DstRect,bitmap);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Process");
				goto EXIT;
			}

			res = MLiveEffect_End(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_End");
				goto EXIT;
			}

			res = MLiveEffect_Destroy(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Destroy");
				goto EXIT;
			}
		}
		if (ProcessRect)
		{
			MMemFree(MNull,ProcessRect);
		}
	}
	else
	{
		goto EXIT;
	}

EXIT:
	if (bitmap)
	{
		MMemFree(MNull,bitmap);
	}

	env->ReleasePrimitiveArrayCritical(imgData,pData,0);
	env->ReleasePrimitiveArrayCritical(jpara,para,0);
	return res;
}

static jint Java_com_arcsoft_arcsoftservice_ArcSoftService_SaveLiveEffect(JNIEnv *env, jobject c, jstring uri_Src, jstring uri_Dst, 
				jint effectID, jintArray jpara,jint dwparamnum,jint quality)
{
	jobject 	obj,obj1;
	jfieldID	fieldId;
	jclass      clazz;
	MBITMAP bm;
	MBITMAP bitmap;
	const char *str1 = (uri_Dst == 0) ? 0 : env->GetStringUTFChars(uri_Dst,0 );
	const char *str2 = (uri_Src == 0) ? 0 : env->GetStringUTFChars(uri_Src,0 );
	MRESULT res = MOK;
	MLong i = 0,lSize = 0;
	MRECT DstRect;
	MRECT tempRect;
	MRECT *ProcessRect;
	MByte *pData;
	MLong lwidth,lheight,linebytes;
	MHandle phLiveEffect;
	SPOTLIGHTPARAM spot_para;
	MOSAICPARAM mosaic_para;
	WarpPinchPARAM warp_para;
	BigEyePARAM    bigeye_para;
	BLURPARAM      blur_para;
	HMSTREAM  stream1;
	int *para = (int *)env->GetPrimitiveArrayCritical(jpara,0);
	
	if (dwparamnum == 0 || (!str1) || (!str2))
	{
		return MERR_INVALID_PARAM;
	}

	stream1 = MStreamOpenFromFile(str2,STREAM_READ);
	MGetImgInfo(stream1,&lwidth,&lheight);
	MStreamClose(stream1);

	linebytes = LINE_BYTES16(lwidth,16);
	pData = (MByte *)MMemAlloc(MNull,linebytes * lheight);
	MMemSet(pData,0,linebytes * lheight);

//	bitmap = (MBITMAP *)MMemAlloc(MNull,sizeof(MBITMAP));
//	if(!bitmap)
//	{
//		log_to_file("MERR_NO_MEMORY exit!");
//		res = MERR_NO_MEMORY;
//		goto EXIT;
//	}
	bitmap.dwPixelArrayFormat = MPAF_RGB16_R5G6B5;
	bitmap.lWidth          = lwidth;
	bitmap.lHeight         = lheight;
	bitmap.lPitch[0]       = linebytes;
	bitmap.lPitch[1]       = 0;
	bitmap.lPitch[2]       = 0;
	bitmap.pPlane[0]       = pData;
	bitmap.pPlane[1]       = MNull;
	bitmap.pPlane[2]       = MNull;

	//decode to rgb565
	res = decodeFile(&bitmap,str2);
	if(res !=MOK)
	{
	    log_to_file("decodeFile exit!");
		goto EXIT;
	}

 	DstRect.left = 0;
 	DstRect.right = lwidth;
 	DstRect.top  = 0;
 	DstRect.bottom = lheight;

	if (effectID == MLIVEEFFECT_ID_SPOTLIGHT)//spotlight
	{
		lSize = sizeof(SPOTLIGHTPARAM);
		ProcessRect = (MRECT *)MMemAlloc(MNull,sizeof(MRECT) * dwparamnum);
	
		for (i = 0; i < dwparamnum;i++)
		{
			ProcessRect[i].left = para[(i << 2)] - para[(i << 2) + 2];
			ProcessRect[i].right = para[(i << 2)] + para[(i << 2) + 2];
			ProcessRect[i].top  = para[(i << 2) + 1] - para[(i << 2) + 3];
			ProcessRect[i].bottom = para[(i << 2) + 1] + para[(i << 2) + 3];
		}
		for (i = 0;i < dwparamnum;i++)
		{
			spot_para.lcircleCentX = para[(i << 2)];
			spot_para.lcircleCentY = para[(i << 2) + 1];
			spot_para.lcircleRadiusx = para[(i << 2) + 2];
			spot_para.lcircleRadiusy = para[(i << 2) + 3];

			tempRect.left = ProcessRect[i].left;
			tempRect.right = ProcessRect[i].right;
			tempRect.top   = ProcessRect[i].top;
			tempRect.bottom = ProcessRect[i].bottom;

			res = MLiveEffect_Create(lwidth,lheight,effectID,&spot_para,lSize,&phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Create");
				goto EXIT;
			}

			res = MLiveEffect_Begin(phLiveEffect, lwidth, lheight, &tempRect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Begin");
				goto EXIT;
			}

			res = MLiveEffect_Process(phLiveEffect,&DstRect,&bitmap,&DstRect,&bitmap);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Process");
				goto EXIT;
			}

			res = MLiveEffect_End(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_End");
				goto EXIT;
			}

			res = MLiveEffect_Destroy(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Destroy");
				goto EXIT;
			}
		}

		multi_area(&bitmap,&bitmap,ProcessRect,dwparamnum);

		if (ProcessRect)
		{
			MMemFree(MNull,ProcessRect);
		}
	}
	else if (effectID == MLIVEEFFECT_ID_MOSAIC)//mosaic
	{
		lSize  = sizeof(MOSAICPARAM);
		for (i = 0;i < dwparamnum;i++)
		{
			tempRect.left  = para[i * 5];
			tempRect.right = para[i * 5 + 1];
			tempRect.top   = para[i * 5 + 2];
			tempRect.bottom= para[i * 5 + 3];
			mosaic_para.lMosaicSize = para[i * 5 + 4];
			
			res = MLiveEffect_Create(lwidth,lheight,effectID,&mosaic_para,lSize,&phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Create");
				goto EXIT;
			}
				
			res = MLiveEffect_Begin(phLiveEffect, lwidth, lheight, &tempRect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Begin");
				goto EXIT;
			}
				
			res = MLiveEffect_Process(phLiveEffect,&DstRect,&bitmap,&DstRect,&bitmap);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Process");
				goto EXIT;
			}
			res = MLiveEffect_End(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_End");
				goto EXIT;
			}
			res = MLiveEffect_Destroy(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Destroy");
				goto EXIT;
			}
		}
	}
	else if (effectID == MLIVEEFFECT_ID_WARP || effectID == MLIVEEFFECT_ID_PINCH)//warp or pinch
	{
		lSize = sizeof(WarpPinchPARAM);
		ProcessRect = (MRECT *)MMemAlloc(MNull,sizeof(MRECT) * dwparamnum);
		for (i = 0; i < dwparamnum;i++)
		{
			ProcessRect[i].left = para[(i * 5 )] - para[(i * 5 ) + 2];
			ProcessRect[i].right = para[(i * 5 )] + para[(i * 5 ) + 2];
			ProcessRect[i].top  = para[(i * 5 ) + 1] - para[(i * 5 ) + 3];
			ProcessRect[i].bottom = para[(i * 5 ) + 1] + para[(i * 5 ) + 3];
		}
		for (i = 0;i < dwparamnum;i++)
		{
			warp_para.lcircleCentX = para[i * 5 ];
			warp_para.lcircleCentY = para[i * 5 + 1];
			warp_para.lcircleRadiusx = para[i * 5 + 2];
			warp_para.lcircleRadiusy = para[i * 5 + 3];
			warp_para.lWarpPinchSize = para[i * 5 + 4];

			tempRect.left = ProcessRect[i].left;
			tempRect.right = ProcessRect[i].right;
			tempRect.top   = ProcessRect[i].top;
			tempRect.bottom = ProcessRect[i].bottom;

			res = MLiveEffect_Create(lwidth,lheight,effectID,&warp_para,lSize,&phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Create");
				goto EXIT;
			}

			res = MLiveEffect_Begin(phLiveEffect, lwidth, lheight, &tempRect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Begin");
				goto EXIT;
			}

			res = MLiveEffect_Process(phLiveEffect,&DstRect,&bitmap,&DstRect,&bitmap);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Process");
				goto EXIT;
			}

			res = MLiveEffect_End(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_End");
				goto EXIT;
			}

			res = MLiveEffect_Destroy(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Destroy");
				goto EXIT;
			}
		}
		if (ProcessRect)
		{
			MMemFree(MNull,ProcessRect);
		}
	}
	else if (effectID == MLIVEEFFECT_ID_BIGEYE)
	{
		lSize = sizeof(BigEyePARAM);
		ProcessRect = (MRECT *)MMemAlloc(MNull,sizeof(MRECT) * dwparamnum);
		for (i = 0; i < dwparamnum;i++)
		{
			ProcessRect[i].left = para[(i * 5 )] - para[(i * 5 ) + 2];
			ProcessRect[i].right = para[(i * 5 )] + para[(i * 5 ) + 2];
			ProcessRect[i].top  = para[(i * 5 ) + 1] - para[(i * 5 ) + 3];
			ProcessRect[i].bottom = para[(i * 5 ) + 1] + para[(i * 5 ) + 3];
		}
		for (i = 0;i < dwparamnum;i++)
		{
			bigeye_para.lcircleCentX = para[i * 5 ];
			bigeye_para.lcircleCentY = para[i * 5 + 1];
			bigeye_para.lcircleRadiusx = para[i * 5 + 2];
			bigeye_para.lcircleRadiusy = para[i * 5 + 3];
			bigeye_para.lBigEyeSize = para[i * 5 + 4];

			tempRect.left = ProcessRect[i].left;
			tempRect.right = ProcessRect[i].right;
			tempRect.top   = ProcessRect[i].top;
			tempRect.bottom = ProcessRect[i].bottom;

			res = MLiveEffect_Create(lwidth,lheight,effectID,&bigeye_para,lSize,&phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Create");
				goto EXIT;
			}

			res = MLiveEffect_Begin(phLiveEffect, lwidth, lheight, &tempRect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Begin");
				goto EXIT;
			}

			res = MLiveEffect_Process(phLiveEffect,&DstRect,&bitmap,&DstRect,&bitmap);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Process");
				goto EXIT;
			}

			res = MLiveEffect_End(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_End");
				goto EXIT;
			}

			res = MLiveEffect_Destroy(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Destroy");
				goto EXIT;
			}
		}
		if (ProcessRect)
		{
			MMemFree(MNull,ProcessRect);
		}
	}
	else if (effectID == MLIVEEFFECT_ID_BLUR)
	{
		lSize = sizeof(BigEyePARAM);
		dwparamnum = 1;
		ProcessRect = (MRECT *)MMemAlloc(MNull,sizeof(MRECT) * dwparamnum);
		for (i = 0; i < dwparamnum;i++)
		{
			ProcessRect[i].left = para[(i * 5 )] - para[(i * 5 ) + 2];
			ProcessRect[i].right = para[(i * 5 )] + para[(i * 5 ) + 2];
			ProcessRect[i].top  = para[(i * 5 ) + 1] - para[(i * 5 ) + 3];
			ProcessRect[i].bottom = para[(i * 5 ) + 1] + para[(i * 5 ) + 3];
		}
		for (i = 0;i < dwparamnum;i++)
		{
			blur_para.lcircleCentX = para[i * 5 ];
			blur_para.lcircleCentY = para[i * 5 + 1];
			blur_para.lcircleRadiusx = para[i * 5 + 2];
			blur_para.lcircleRadiusy = para[i * 5 + 3];
			blur_para.lBlurSize = para[i * 5 + 4];

			tempRect.left = ProcessRect[i].left;
			tempRect.right = ProcessRect[i].right;
			tempRect.top   = ProcessRect[i].top;
			tempRect.bottom = ProcessRect[i].bottom;

			res = MLiveEffect_Create(lwidth,lheight,effectID,&blur_para,lSize,&phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Create");
				goto EXIT;
			}

			res = MLiveEffect_Begin(phLiveEffect, lwidth, lheight, &DstRect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Begin");
				goto EXIT;
			}

			res = MLiveEffect_Process(phLiveEffect,&DstRect,&bitmap,&DstRect,&bitmap);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Process");
				goto EXIT;
			}

			res = MLiveEffect_End(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_End");
				goto EXIT;
			}

			res = MLiveEffect_Destroy(phLiveEffect);
			if(res !=MOK)
			{
				log_to_file("MLiveEffect_Destroy");
				goto EXIT;
			}
		}
		if (ProcessRect)
		{
			MMemFree(MNull,ProcessRect);
		}
	}
	else
	{
		goto EXIT;
	}

	//encode
	res = encodeFile(str1,&bitmap,MNull,MNull,quality);
	if(res !=MOK)
		goto EXIT;

EXIT:
	if (pData)
	{
		MMemFree(MNull,pData);
	}

	env->ReleasePrimitiveArrayCritical(jpara,para,0);
	env->ReleaseStringUTFChars(uri_Dst, str1);
	env->ReleaseStringUTFChars(uri_Src, str2);
	return res;
}

static jint JNICALL Java_com_arcsoft_arcsoftservice_ArcSoftService_MergeEffect
  (JNIEnv *env, jobject c, jbyteArray imgData, jbyteArray imgforeData, jbyteArray maskData, jint backwidth, jint backlinebyte,jint backheight,
   jint forewidth, jint forelinebyte, jint foreheight,jint maskwidth, jint masklinebyte, jint maskheight,
  jint Opacity, jint offsetX, jint offsetY)
{
    int res = 0;
	MBITMAP *bitmap_back,*bitmap_fore,*bitmap_mask;
	MPOINT point;
	int *pData = (int *)env->GetPrimitiveArrayCritical(imgData,0);
	int *pData2 = (int *)env->GetPrimitiveArrayCritical(imgforeData,0);
	int *pData3 = (int *)env->GetPrimitiveArrayCritical(maskData,0);

	point.x = offsetX;
	point.y = offsetY;

	bitmap_back = (MBITMAP *)MMemAlloc(MNull,sizeof(MBITMAP));
	if(!bitmap_back)
	{
		res = MERR_NO_MEMORY;
		goto EXIT;
	}
	bitmap_back->dwPixelArrayFormat = MPAF_RGB16_R5G6B5;
	bitmap_back->lWidth          = backwidth;
	bitmap_back->lHeight         = backheight;
	bitmap_back->lPitch[0]       = backlinebyte;
	bitmap_back->lPitch[1]       = 0;
	bitmap_back->lPitch[2]       = 0;
	bitmap_back->pPlane[0]       = (MByte *)pData;
	bitmap_back->pPlane[1]       = MNull;
	bitmap_back->pPlane[2]       = MNull;

	bitmap_fore = (MBITMAP *)MMemAlloc(MNull,sizeof(MBITMAP));
	if(!bitmap_fore)
	{
		res = MERR_NO_MEMORY;
		goto EXIT;
	}
	bitmap_fore->dwPixelArrayFormat = MPAF_RGB16_R5G6B5;
	bitmap_fore->lWidth          = forewidth;
	bitmap_fore->lHeight         = foreheight;
	bitmap_fore->lPitch[0]       = forelinebyte;
	bitmap_fore->lPitch[1]       = 0;
	bitmap_fore->lPitch[2]       = 0;
	bitmap_fore->pPlane[0]       = (MByte *)pData2;
	bitmap_fore->pPlane[1]       = MNull;
	bitmap_fore->pPlane[2]       = MNull;

	bitmap_mask = (MBITMAP *)MMemAlloc(MNull,sizeof(MBITMAP));
	if(!bitmap_mask)
	{
		res = MERR_NO_MEMORY;
		goto EXIT;
	}
	bitmap_mask->dwPixelArrayFormat = MPAF_GRAY8;
	bitmap_mask->lWidth          = maskwidth;
	bitmap_mask->lHeight         = maskheight;
	bitmap_mask->lPitch[0]       = masklinebyte;
	bitmap_mask->lPitch[1]       = 0;
	bitmap_mask->lPitch[2]       = 0;
	bitmap_mask->pPlane[0]       = (MByte *)pData3;
	bitmap_mask->pPlane[1]       = MNull;
	bitmap_mask->pPlane[2]       = MNull;

	MdBitmapMerge_NEW(bitmap_back,bitmap_fore,&point,bitmap_mask,&point,Opacity);

EXIT:

		env->ReleasePrimitiveArrayCritical(imgData,pData,0);
		env->ReleasePrimitiveArrayCritical(imgforeData,pData2,0);
		env->ReleasePrimitiveArrayCritical(maskData,pData3,0);

	if (bitmap_back)
	{
		MMemFree(MNull,bitmap_back);
	}

	if (bitmap_fore)
	{
		MMemFree(MNull,bitmap_fore);
	}

	if (bitmap_mask)
	{
		MMemFree(MNull,bitmap_mask);
	}
	return res;
}

static jint Java_com_arcsoft_arcsoftservice_ArcSoftService_SaveMergeEffect(JNIEnv *env, jobject c,
		jstring Dstfilename,jstring backfilename, jstring forefilename, jstring maskfilename, jint Opacity, jint Rect_left,
			jint Rect_right, jint Rect_top, jint Rect_bottom ,jint quality)
{
	const char *str_back = (backfilename == 0) ? 0 : env->GetStringUTFChars(backfilename,0 );
	const char *str_fore = (forefilename == 0) ? 0 : env->GetStringUTFChars(forefilename,0 );
	const char *str_mask = (maskfilename == 0) ? 0 : env->GetStringUTFChars(maskfilename,0 );
	const char *str_Dst = (Dstfilename == 0) ? 0 : env->GetStringUTFChars(Dstfilename,0 );
//	int quality = 100;
	int res = 0;
	SkBitmap bm_back,bm_fore,bm_mask;
	MBITMAP *bitmap_back,*bitmap_fore,*bitmap_mask;
	MRECT DstRect;
	MByte *pbackData,*pforedata,*pmaskdata;
	MByte *pMask;
	MHandle phLiveEffect;
	SPOTLIGHTPARAM spot_para;
	MLong lWidth = 0,lHeight = 0;
	int lSize = 0;
	MPOINT point;

	//lWidth = Rect_right - Rect_left;
	//lHeight = Rect_bottom - Rect_top;

	point.x = 0;//Rect_left;
	point.y = 0;//Rect_top;
	//decode to rgb565
//	res = decodeFile(&bm_back,str_back,SkBitmap::kRGB_565_Config);
	if(res !=MOK)
		goto EXIT;
	pbackData = (MByte *)bm_back.getPixels();

//	res = decodeFile(&bm_fore,str_fore,SkBitmap::kRGB_565_Config);
	if(res !=MOK)
		goto EXIT;
	pforedata = (MByte *)bm_fore.getPixels();

//	res = decodeFile(&bm_mask,str_mask,SkBitmap::kA8_Config);//skia jpg not support A8,change to ARGB888
	if(res !=MOK)
		goto EXIT;
	pmaskdata = (MByte *)bm_mask.getPixels();

	bitmap_back = (MBITMAP *)MMemAlloc(MNull,sizeof(MBITMAP));
	if(!bitmap_back)
	{
		res = MERR_NO_MEMORY;
		goto EXIT;
	}
	bitmap_back->dwPixelArrayFormat = MPAF_RGB16_R5G6B5;
	bitmap_back->lWidth          = bm_back.width();
	bitmap_back->lHeight         = bm_back.height();
	bitmap_back->lPitch[0]       = bm_back.rowBytes();
	bitmap_back->lPitch[1]       = 0;
	bitmap_back->lPitch[2]       = 0;
	bitmap_back->pPlane[0]       = (MByte *)pbackData;
	bitmap_back->pPlane[1]       = MNull;
	bitmap_back->pPlane[2]       = MNull;

	bitmap_fore = (MBITMAP *)MMemAlloc(MNull,sizeof(MBITMAP));
	if(!bitmap_fore)
	{
		res = MERR_NO_MEMORY;
		goto EXIT;
	}
	bitmap_fore->dwPixelArrayFormat = MPAF_RGB16_R5G6B5;
	bitmap_fore->lWidth          = bm_fore.width();
	bitmap_fore->lHeight         = bm_fore.height();
	bitmap_fore->lPitch[0]       = bm_fore.rowBytes();
	bitmap_fore->lPitch[1]       = 0;
	bitmap_fore->lPitch[2]       = 0;
	bitmap_fore->pPlane[0]       = (MByte *)pforedata;
	bitmap_fore->pPlane[1]       = MNull;
	bitmap_fore->pPlane[2]       = MNull;

	pMask = (MByte *)MMemAlloc(MNull,bm_mask.width() * bm_mask.height() * sizeof(MByte));
	if(!pMask)
	{
		res = MERR_NO_MEMORY;
		goto EXIT;
	}

	GetAlpha(pMask,pmaskdata,bm_mask.width(),bm_mask.height(),bm_mask.width(),bm_mask.rowBytes());
	
	bitmap_mask = (MBITMAP *)MMemAlloc(MNull,sizeof(MBITMAP));
	if(!bitmap_mask)
	{
		res = MERR_NO_MEMORY;
		goto EXIT;
	}
	bitmap_mask->dwPixelArrayFormat = MPAF_GRAY8;
	bitmap_mask->lWidth          = bm_mask.width();
	bitmap_mask->lHeight         = bm_mask.height();
	bitmap_mask->lPitch[0]       = bm_mask.width();//bm_mask.rowBytes();
	bitmap_mask->lPitch[1]       = 0;
	bitmap_mask->lPitch[2]       = 0;
	bitmap_mask->pPlane[0]       = (MByte *)pMask;//pmaskdata;
	bitmap_mask->pPlane[1]       = MNull;
	bitmap_mask->pPlane[2]       = MNull;

	//do effect
	MdBitmapMerge_NEW(bitmap_back,bitmap_fore,&point,bitmap_mask,&point,Opacity);

	//encode
//	res = encodeFile(str_Dst,bm_back,SkImageEncoder::kJPEG_Type,100);
	if(res != MOK)
		goto EXIT;

EXIT:
	if (bitmap_back)
	{
		MMemFree(MNull,bitmap_back);
	}

	if (bitmap_fore)
	{
		MMemFree(MNull,bitmap_fore);
	}

	if (bitmap_mask)
	{
		MMemFree(MNull,bitmap_mask);
	}
	if(pMask)
	{
		MMemFree(MNull,pMask);
	}

	env->ReleaseStringUTFChars(backfilename, str_back);
	env->ReleaseStringUTFChars(forefilename, str_fore);
	env->ReleaseStringUTFChars(maskfilename, str_mask);
	env->ReleaseStringUTFChars(Dstfilename, str_Dst);


	return res;
}


static jint Java_com_arcsoft_arcsoftservice_ArcSoftService_rerjpeg(JNIEnv *env, jobject thiz,
		jstring input_filename,jstring output_filename,jint quality)
{
	const char *str_src = (input_filename == 0) ? 0 : env->GetStringUTFChars(input_filename,0 );
	const char *str_Dst = (output_filename == 0) ? 0 : env->GetStringUTFChars(output_filename,0 );
	MBITMAP bitmap;
	MLong lWidth = 0,lHeight = 0,linebytes = 0;
	int lSize = 0;
	MByte * pSrcData;
 	int redeye_num, result;
	HMSTREAM stream1 = MNull;
	MByte *pData = MNull;

	stream1 = MStreamOpenFromFile(str_src,STREAM_READ);
	MGetImgInfo(stream1,&lWidth,&lHeight);
	MStreamClose(stream1);

	linebytes = LINE_BYTES16(lWidth,16);
	pData = (MByte *)MMemAlloc(MNull,linebytes * lHeight);
	MMemSet(pData,0,linebytes * lHeight);

	bitmap.dwPixelArrayFormat = MPAF_RGB16_R5G6B5;
	bitmap.lWidth          = lWidth;
	bitmap.lHeight         = lHeight;
	bitmap.lPitch[0]       = linebytes;
	bitmap.lPitch[1]       = 0;
	bitmap.lPitch[2]       = 0;
	bitmap.pPlane[0]       = pData;
	bitmap.pPlane[1]       = MNull;
	bitmap.pPlane[2]       = MNull;

	//decode to rgb565
	result = decodeFile(&bitmap,str_src);
	if(result !=MOK)
	{
		env->ReleaseStringUTFChars(input_filename, str_src);
		env->ReleaseStringUTFChars(output_filename, str_Dst);
		return result;// no jpeg
	}

 	redeye_num = Do_RER(pSrcData, lWidth, lHeight, linebytes);

// 	 log_to_file("bm_src.width() = %d",bm_src.width());
// 	 log_to_file("bm_src.height() = %d",bm_src.height());
// 	 log_to_file("bm_src.rowBytes() = %d",bm_src.rowBytes());

 //  dump_to_file("/data/jichen/rer.dat", (uint8_t *)pSrcData,  bm_src.width()*2*bm_src.height());

	//encode
//	if(encodeFile(str_Dst,bm_src,SkImageEncoder::kJPEG_Type,100) != MOK)
//       log_to_file("encodeFile Failed");
//	else
//		log_to_file("encodeFile OK");

	encodeFile(str_Dst,&bitmap,MNull,MNull,85);

	env->ReleaseStringUTFChars(input_filename, str_src);
	env->ReleaseStringUTFChars(output_filename, str_Dst);

	if (pData)
	{
		MMemFree(MNull,pData);
	}

    return redeye_num;
}


static jint Java_com_arcsoft_arcsoftservice_ArcSoftService_fbjpeg(JNIEnv *env, jobject thiz,
		jstring input_filename,jstring output_filename, jint option, jint cleanLevel, jintArray FaceRes, 
		jint FaceNum,jint quality)
{
	const char *str_src = (input_filename == 0) ? 0 : env->GetStringUTFChars(input_filename,0 );
	const char *str_Dst = (output_filename == 0) ? 0 : env->GetStringUTFChars(output_filename,0 );
	int *pFaceRes = (int *)env->GetPrimitiveArrayCritical(FaceRes,0);
	MBITMAP bitmap;
	MLong lWidth = 0,lHeight = 0,linebytes = 0;
	int lSize = 0;
	MByte * pSrcData;
 	int result;
	HMSTREAM stream1 = MNull;
	MByte *pData = MNull;

	stream1 = MStreamOpenFromFile(str_src,STREAM_READ);
	MGetImgInfo(stream1,&lWidth,&lHeight);
	MStreamClose(stream1);

	linebytes = LINE_BYTES16(lWidth,16);
	pData = (MByte *)MMemAlloc(MNull,linebytes * lHeight);
	MMemSet(pData,0,linebytes * lHeight);

	bitmap.dwPixelArrayFormat = MPAF_RGB16_R5G6B5;
	bitmap.lWidth          = lWidth;
	bitmap.lHeight         = lHeight;
	bitmap.lPitch[0]       = linebytes;
	bitmap.lPitch[1]       = 0;
	bitmap.lPitch[2]       = 0;
	bitmap.pPlane[0]       = pData;
	bitmap.pPlane[1]       = MNull;
	bitmap.pPlane[2]       = MNull;

	//decode to rgb565
	result = decodeFile(&bitmap,str_src);
	if(result !=MOK)
	{
		env->ReleasePrimitiveArrayCritical(FaceRes,pFaceRes,0);
		env->ReleaseStringUTFChars(input_filename, str_src);
		env->ReleaseStringUTFChars(output_filename, str_Dst);	
		return result;
	}

	result = do_FB(pData ,  lWidth, lHeight,linebytes, 
				   option,  cleanLevel, pFaceRes, FaceNum);
  
 //	 log_to_file("Java_com_arcsoft_arcsoftservice_ArcSoftService_fbjpeg bm_src.width() = %d",bm_src.width());
 //	 log_to_file("bm_src.height() = %d",bm_src.height());
 //	 log_to_file("bm_src.rowBytes() = %d",bm_src.rowBytes());

   //dump_to_file("/data/jichen/rer.dat", (uint8_t *)pSrcData,  bm_src.width()*2*bm_src.height());

	//encode
	encodeFile(str_Dst,&bitmap,MNull,MNull,85);

	env->ReleasePrimitiveArrayCritical(FaceRes,pFaceRes,0);
	env->ReleaseStringUTFChars(input_filename, str_src);
	env->ReleaseStringUTFChars(output_filename, str_Dst);

	if (pData)
	{
		MMemFree(MNull,pData);
	}

    return result;
}

static const char *classPathName = "com/arcsoft/arcsoftservice/ArcSoftService";

static JNINativeMethod methods[] = {
  {"FaceDetect", "([BIIIII[I[I)I", (void*)Java_com_arcsoft_arcsoftservice_ArcSoftService_FaceDetect },
  {"FaceRegister", "([BIIII[I[B[I)I", (void*)Java_com_arcsoft_arcsoftservice_ArcSoftService_FaceRegister },
  {"FaceRegonition", "([BI[B[I[I[I)I", (void*)Java_com_arcsoft_arcsoftservice_ArcSoftService_FaceRegonition },
  {"RER", "([BIII)I", (void*)Java_com_arcsoft_arcsoftservice_ArcSoftService_RER },
  {"FB", "([BIIIII[II)I", (void*)Java_com_arcsoft_arcsoftservice_ArcSoftService_FB },
 {"LiveEffect","([BIIII[II)I",(void*)Java_com_arcsoft_arcsoftservice_ArcSoftService_LiveEffect},
  {"SaveLiveEffect","(Ljava/lang/String;Ljava/lang/String;I[III)I",(void*)Java_com_arcsoft_arcsoftservice_ArcSoftService_SaveLiveEffect},
  {"MergeEffect","([B[B[BIIIIIIIIIIII)I",(void*)Java_com_arcsoft_arcsoftservice_ArcSoftService_MergeEffect},
  {"SaveMergeEffect","(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;IIIIII)I",(void*)Java_com_arcsoft_arcsoftservice_ArcSoftService_SaveMergeEffect},  
  {"rerjpeg", "(Ljava/lang/String;Ljava/lang/String;I)I", (void*)Java_com_arcsoft_arcsoftservice_ArcSoftService_rerjpeg },
  {"fbjpeg", "(Ljava/lang/String;Ljava/lang/String;II[III)I", (void*)Java_com_arcsoft_arcsoftservice_ArcSoftService_fbjpeg },
};


/*
 * Register several native methods for one class.
 */
static int registerNativeMethods(JNIEnv* env, const char* className,
    JNINativeMethod* gMethods, int numMethods)
{
    jclass clazz;

    clazz = env->FindClass(className);
    if (clazz == NULL) {
        log_to_file("Native registration unable to find class '%s'", className);
        return JNI_FALSE;
    }

//	if (AndroidRuntime::registerNativeMethods(env, className, methods,  NELEM(methods))) < 0) {
     if (env->RegisterNatives(clazz, gMethods, numMethods) < 0) {
        log_to_file("RegisterNatives failed for '%s'", className);
        return JNI_FALSE;
    }

  
    return JNI_TRUE;
}

/*
 * Register native methods for all classes we know about.
 *
 * returns JNI_TRUE on success.
 */
static int registerNatives(JNIEnv* env)
{
  if (!registerNativeMethods(env, classPathName,
                 methods, sizeof(methods) / sizeof(methods[0]))) {
	  log_to_file("registerNativeMethods failed\n");
    return JNI_FALSE;
  }

  return JNI_TRUE;
}


// ----------------------------------------------------------------------------

/*
 * This is called by the VM when the shared library is first loaded.
 */
 
typedef union {
    JNIEnv* env;
    void* venv;
} UnionJNIEnvToVoid;

jint JNI_OnLoad(JavaVM* vm, void* reserved)
{
    UnionJNIEnvToVoid uenv;
    uenv.venv = NULL;
    jint result = -1;
    JNIEnv* env = NULL;
    
    log_to_file("JNI_OnLoad");

    if (vm->GetEnv(&uenv.venv, JNI_VERSION_1_4) != JNI_OK) {
        log_to_file("ERROR: GetEnv failed");
        goto bail;
    }
    env = uenv.env;

    if (registerNatives(env) != JNI_TRUE) {
        log_to_file("ERROR: registerNatives failed");
        goto bail;
    }
    
    result = JNI_VERSION_1_4;
    
bail:
    return result;
}
